#include "include/mns.h"
#include "include/inetraw.h"

#define pcap_ip(a) ((struct ip *)(p + (a)))

void pcap_alarm(int sigio);

struct in_addr local_ip; /* local ip address */
struct in_addr target_ip; /* target ip address */

u_short src_port; /* source port */
u_short dst_port; /* destination port */

int clicks;

int pcap_off; /* header offset for pcap_next captured packets */

pcap_t *pd_raw; /* packet capture descriptor */
int sd_raw; /* raw socket descriptor */
int sd_icmp; /* icmp socket descriptor */

char *init_packet_devices(int proto) {
   char errbuf[PCAP_ERRBUF_SIZE];
   char *e_msg;

   pcap_off = 0;
   pd_raw = pcap_open_live(device, MIN_PKT << 2, 0, MIN_PKT, errbuf);
   if(pd_raw == NULL) {
      e_msg = malloc(sizeof("\n\t- opening pcap device: ") + strlen(errbuf));
      sprintf(e_msg, "\n\t- opening pcap device: %s", errbuf);
      return e_msg;
   }

   if((sd_raw = socket(AF_INET, SOCK_RAW, proto)) < 0) {
      e_msg = malloc(sizeof("\n\t- opening raw socket: ") +
                     strlen(strerror(errno)));
      sprintf(e_msg, "\n\t- opening raw socket: %s", strerror(errno));
      return e_msg;
   }

   if((sd_icmp = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP)) < 0) {
      e_msg = malloc(sizeof("\n\t- opening raw socket: ") +
                     strlen(strerror(errno)));
      sprintf(e_msg, "\n\t- opening raw socket: %s", strerror(errno));
      return e_msg;
   }
   
   return NULL;
}

void term_packet_devices(void) {
   close(sd_raw);
   close(sd_icmp);
   pcap_close(pd_raw);
}

void pcap_alarm(int sigio) {
   SAin sa_dest;
   struct icmp_h icmp;

   sa_dest.sin_family = AF_INET;
   sa_dest.sin_addr = target_ip;
   sa_dest.sin_port = htons((rand() % 1000) + 1000);

   bzero(&icmp, sizeof(icmp));
   icmp.icmp_type = ICMP_ECHO;
   icmp.icmp_un.echo.id = getpid();
   icmp.icmp_sum = in_cksum((u_short *)&icmp, sizeof(icmp));
   
   sendto(sd_icmp, &icmp, sizeof(icmp), 0, (SA *)&sa_dest, sizeof(SAin));
   
   if(++clicks == 5) {
      printf("cannot receive messages on the interface '%s'\n", device);
      exit(EXIT_FAILURE);
   }
   
   alarm(1);
}

void sendtcp(hdrinfo *routing, u_short flags) {
  struct tcphdr tcp;
  SAin sa_dest;

  tcp.th_sport = src_port;
  tcp.th_dport = dst_port;
  tcp.th_seq = htonl (routing->seq);
  tcp.th_ack = 0;
  tcp.th_x2 = 0;
  tcp.th_off = 5;
  tcp.th_flags = flags;
  tcp.th_win = htons(0x1000 | (rand () % 0x1000));
  tcp.th_urp = 0;
  tcp.th_sum = 0;
/*-- TCP Checksum --*/
#ifdef SOLARIS_CKSUM_BUG
  tcp.th_sum = sizeof (struct tcphdr);
#else
  tcp.th_sum = tcp_cksum((char *)&tcp, sizeof(tcp), local_ip, target_ip);
#endif /* SOLARIS_CKSUM_BUG */

   sa_dest.sin_family = AF_INET;
   sa_dest.sin_addr = target_ip;
   sa_dest.sin_port = 0;
   
   if(sendto(sd_raw, &tcp, sizeof(tcp), 0, (SA *)&sa_dest, sizeof(SAin)) < 0)
      perror("sending message");
}

u_long get_packet(struct ip *ip_ptr, char *packet, int pkt_len, int side) {
   struct ip ip2;
   struct pcap_pkthdr pcap_hdr;
   char *p = NULL;
   int i, max;

   signal(SIGALRM, &pcap_alarm);
   clicks = 0;

   if(ip_ptr == NULL) ip_ptr = &ip2;

   alarm(1);
   if((p = (char *)pcap_next(pd_raw, &pcap_hdr)) == NULL) return 0;
   alarm(0);

   if(!pcap_off) max = min(50, pcap_hdr.caplen);
   else max = pcap_off + 1;
   
   for(i = pcap_off; i < max; i++) {
      memcpy(ip_ptr, pcap_ip(i), sizeof(struct ip));

      ip_ptr->ip_sum = 0;
      ip_ptr->ip_sum = in_cksum((u_short *)ip_ptr, sizeof(struct ip));

      if(pcap_ip(i)->ip_sum == ip_ptr->ip_sum) {
         if( (side == SRC ?
             pcap_ip(i)->ip_src.s_addr :
             pcap_ip(i)->ip_dst.s_addr) == target_ip.s_addr) {

            bzero(packet, pkt_len);
            memcpy(packet, p + i + sizeof(struct ip),
                   min(pkt_len, pcap_hdr.caplen));
                   
            if(!pcap_off) pcap_off = i;
            return  (side == DST ?
             pcap_ip(i)->ip_src.s_addr :
             pcap_ip(i)->ip_dst.s_addr);
         } else break;
      }
   }
   
   return 0;
}

void sendudp(int sd_udp) {
  struct udp_h udp;
  SAin sa_dest;

   udp.udp_sport = 0;
   udp.udp_dport = dst_port;
   udp.udp_len = htons(8);
   udp.udp_sum = 0;

   sa_dest.sin_family = AF_INET;
   sa_dest.sin_addr = target_ip;
   sa_dest.sin_port = dst_port;
   
   if(sendto(sd_udp, &udp, sizeof(udp), 0, (SA *)&sa_dest, sizeof(SAin)) < 0)
      perror("sending message");
}
